package main

import (
	"sync"
	"time"
)

func main() {
	m := NewLockManager()
	m.Acquire(1, Shared, true)
	m.Acquire(1, Shared, true)
	go func() {
		time.Sleep(0.2e9)
		m.Release(1, Shared)
		time.Sleep(0.2e9)
		m.Release(1, Shared)
		time.Sleep(0.5e9)
	}()
	m.Acquire(1, Exclusive, true)
}

type LockMode int

type Lock struct {
	mutex  sync.Mutex
	mode   LockMode
	locks  *[NumModes]chan bool
	unlock chan LockMode
}

type LockManager struct {
	mutex sync.Mutex
	locks map[interface{}]*Lock
}

func NewLockManager() *LockManager {
	return &LockManager{locks: make(map[interface{}]*Lock)}
}

func (m *LockManager) getLock(item interface{}) *Lock {
	m.mutex.Lock()
	l := m.locks[item]
	if l == nil {
		l = &Lock{}
		m.locks[item] = l
	}
	m.mutex.Unlock()
	return l
}

func (m *LockManager) Acquire(item interface{}, mode LockMode, wait bool) bool {
	return m.getLock(item).Acquire(mode, wait)
}

func (m *LockManager) Release(item interface{}, mode LockMode) {
	m.getLock(item).Release(mode)
}

func (l *Lock) Acquire(mode LockMode, wait bool) bool {
again:
	l.mutex.Lock()
	// simple case - we don't need a mediating
	// process when mode was None
	if l.mode == None {
		l.mode = mode
		l.mutex.Unlock()
		return true
	}
	// if there's no locker process, start one.
	if l.locks == nil {
		l.locks = new([NumModes]chan bool)
		for i := range *l.locks {
			l.locks[i] = make(chan bool)
		}
		l.unlock = make(chan LockMode)
		go l.locker(l.mode)
	}
	lockc := l.locks[mode]
	l.mutex.Unlock()

	var locked bool
	if wait {
		locked = <-lockc
	} else {
		var ok bool
		if locked, ok = <-lockc; !ok {
			return false
		}
	}
	// we'll only ever get false if the locker has gone away,
	// in which case we try again.
	if !locked {
		goto again
	}
	return true
}

func (l *Lock) Release(mode LockMode) {
	l.unlock <- mode
}

// set okLocks to the set of locking operations currently available
// from mode.
func setChans(mode LockMode, okLocks, locks *[NumModes]chan bool) {
	for i, ok := range compatibilityMatrix[mode] {
		if ok {
			okLocks[i] = locks[i]
		} else {
			okLocks[i] = nil
		}
	}
}

var freeLocker = make(chan bool)

// locker mediates access to a single Lock.
func (l *Lock) locker(mode LockMode) {
	var okLocks [NumModes]chan bool
	var counts [NumModes]int
	counts[mode] = 1
	setChans(mode, &okLocks, l.locks)
	var freec chan bool
	for {
		lockReq := true
		var newMode LockMode
		select {
		case okLocks[IntentionShared] <- true:
			newMode = IntentionShared
		case okLocks[IntentionExclusive] <- true:
			newMode = IntentionExclusive
		case okLocks[Shared] <- true:
			newMode = Shared
		case okLocks[SharedIntentionExclusive] <- true:
			newMode = SharedIntentionExclusive
		case okLocks[Update] <- true:
			newMode = Update
		case okLocks[IntentionShared] <- true:
			newMode = IntentionShared
		case okLocks[Exclusive] <- true:
			newMode = Exclusive

		case <-freec:
			// we only allow this to happen when there are
			// no locks currently held, so the next request
			// cannot be a Release; if it's a Release,
			// the reader will detect that the locker has gone
			// away and try again.
			for i := range *l.locks {
				close(l.locks[i])
			}
			l.mutex.Lock()
			l.locks = nil
			l.mode = None
			l.mutex.Unlock()
			return

		case m := <-l.unlock:
			counts[m]--
			// move to highest mode with count>0
			var i LockMode
			for i = NumModes - 1; i > 0; i-- {
				if counts[i] > 0 {
					break
				}
			}
			// lock mode is None - make outselves available to be freed.
			if i <= 0 {
				freec = freeLocker
			}
			lockReq = false
			newMode = i
		}
		if lockReq {
			counts[newMode]++
			newMode = mode.maximumOf(newMode)
			freec = nil // not available to be freed
		}
		if newMode != mode {
			mode = newMode
			setChans(mode, &okLocks, l.locks)
		}
	}
}

const (
	None = LockMode(iota)
	IntentionShared
	IntentionExclusive
	Shared
	SharedIntentionExclusive
	Update
	Exclusive
	NumModes
)

var modeNames = [NumModes]string{
	"None",
	"IntentionShared",
	"IntentionExclusive",
	"Shared",
	"SharedIntentionExclusive",
	"Update",
	"Exclusive",
}

var compatibilityMatrix = [NumModes][NumModes]bool{
	[NumModes]bool{true, true, true, true, true, true, true},
	[NumModes]bool{true, true, true, true, true, false, false},
	[NumModes]bool{true, true, true, false, false, false, false},
	[NumModes]bool{true, true, false, true, false, true, false},
	[NumModes]bool{true, true, false, false, false, false, false},
	[NumModes]bool{true, false, false, true, false, false, false},
	[NumModes]bool{true, false, false, false, false, false, false},
}

var conversionMatrix = [][]LockMode{
	[]LockMode{None, IntentionShared, IntentionExclusive, Shared,
		SharedIntentionExclusive, Update, Exclusive},
	[]LockMode{IntentionShared, IntentionShared, IntentionExclusive, Shared,
		SharedIntentionExclusive, Update, Exclusive},
	[]LockMode{IntentionExclusive, IntentionExclusive, IntentionExclusive,
		SharedIntentionExclusive, SharedIntentionExclusive,
		Exclusive, Exclusive},
	[]LockMode{Shared, Shared, SharedIntentionExclusive, Shared,
		SharedIntentionExclusive, Update, Exclusive},
	[]LockMode{SharedIntentionExclusive, SharedIntentionExclusive,
		SharedIntentionExclusive, SharedIntentionExclusive,
		SharedIntentionExclusive, SharedIntentionExclusive,
		Exclusive},
	[]LockMode{Update, Update, Exclusive, Update, SharedIntentionExclusive,
		Update, Exclusive},
	[]LockMode{Exclusive, Exclusive, Exclusive, Exclusive, Exclusive, Exclusive,
		Exclusive},
}

func (m LockMode) isCompatible(mode LockMode) bool {
	return compatibilityMatrix[m][mode]
}

func (m LockMode) maximumOf(mode LockMode) LockMode {
	return conversionMatrix[m][mode]
}

func (m LockMode) String() string {
	return modeNames[m]
}
